C 2D Spectral Poisson solver : x-Fourier y-Chebyshev (assumes homogeneous BC)
C
C  Double precision version.  Requires mylapack.f and myfftD.f
C
      PROGRAM POISS2D
      IMPLICIT NONE
      
      INTEGER size,sizep1,lwork,wsize
      PARAMETER(size=64,sizep1=size+1,lwork=10*size,wsize=size+20)
      REAL*8 pi
      PARAMETER(pi=3.14159265358979323846264338327950288D+0)
      
      INTEGER  n,m,mm1,i,j,k
      REAL*8   xlen,dx,t1,t2
      REAL*8   x(0:size),y(0:size),D1(0:size,0:size),D2(0:size,0:size)
      REAL*8   tu(0:size,0:size),u(0:size,0:size),rhs(0:size,0:size)

C Fourier variables for VFFTPK

      REAL*8 fm(size),wsave(wsize),fftwork(0:size,0:size)
      
C eigen-deomposition variables

      INTEGER  info,prinfo,ipiv(size)
      REAL*8 vl,work(lwork),sum
      REAL*8 D2int(size,size),wr(size),wi(size)
      REAL*8 vr(size,size),vrinv(size,size)
     
      WRITE(6,*) 'Enter n (x) and m (y):'
      READ(5,*) n,m
      mm1=m-1
C      WRITE(6,*) 'Enter x-length:'
C      READ(5,*)  xlen
C to test solver take xlen=2*pi

      xlen=2.0D+0*pi
      
      IF ((size .LT. m) .OR. (size .LT. n)) THEN
         WRITE(6,*) 'SIZE in CHEBY is too small!!'
         STOP
      END IF
C
C find Chebyshev points y, along with D1 & D2
C      
      CALL CHEB(y,D1,D2,m,size)
C
C extract interior of D2
C
      DO j=1,mm1
         DO i=1,mm1
            D2int(i,j)=D2(i,j)
         END DO
      END DO
C
C compute eigendecomposition of D2int
C
      CALL DGEEV('N','V',mm1,D2int,size,wr,wi,vl,1,vr,size,
     &            work,lwork,info)
C
C find inverse of eigenvector matrix vr
C dgetrf does LU, then dgetri finds inverse using it
C
      DO j=1,mm1
         DO i=1,mm1
            vrinv(i,j)=vr(i,j)
         END DO
      END DO

      CALL DGETRF(mm1,mm1,vrinv,size,ipiv,info)
      CALL DGETRI(mm1,vrinv,size,ipiv,work,lwork,info)

C setup x, Fourier multiples, and initialize wsave for FFT calls
C nth Fourier multiple is set arbitrarily to 1 since we will 
C ignore this mode. (see documentation for VRFFTF/VRFFTB: the 
C last modedoes not fit the proper pattern, but we can ignore it 
C since it is a high mode)
C
      dx=xlen/DBLE(n)
      DO j=0,n
         x(j)=DBLE(j)*dx
      END DO
      fm(1)=0.0D+0
      DO J=2,n-2,2
         fm(j)=-(pi*DBLE(j)/xlen)**2
         fm(j+1)=fm(j)
      END DO
      fm(n)=1.0D+0
      
      CALL VRFFTI(n,wsave)
C      
C  intialize tu (true u) and rhs for Poisson solve
C
      DO j=0,n
         DO i=0,m
            tu(i,j)=(DCOS(2.0D+0*x(j))+5.0D+0)*(1.0D+0-y(i)**2)
            rhs(i,j)=-4.0D+0*DCOS(2.0D+0*x(j))*(1.0D+0-y(i)**2)
     &               -2.0D+0*(DCOS(2.0D+0*x(j))+5.0D+0)
         END DO
      END DO
C
C solve Poisson eqn: first FFT periodic rows of rhs.  We take the 
C j=1:n entries.  We only need i=1:m-1 since we have homogeneous 
C  Dirichlet BC            
C      
      CALL VRFFTF(mm1,n,rhs(1,1),fftwork,sizep1,wsave)
C
C multiply by vrinv on left (i=1:mm1,j=1:n) and divide by Fourier/Chebyshev
C multiple, and use fftwork to store result.     
C
      DO j=1,n
         DO i=1,mm1
            sum=0.0D+0
            DO k=1,mm1
               sum=sum+vrinv(i,k)*rhs(k,j)
            END DO
            fftwork(i,j)=sum/(wr(i)+fm(j))
         END DO
      END DO
C      
C multiply by vr on left (i=1:mm1,j=1:n-1) and store result in u
C we only go to n-1 since last coeff will be set to 0
C
      DO j=1,n-1
         DO i=1,mm1
            sum=0.0D+0
            DO k=1,mm1
               sum=sum+vr(i,k)*fftwork(k,j)
            END DO
            u(i,j)=sum
         END DO
      END DO   
C
C set nth Fourier coeff to 0 and transform back
C 
      DO i=1,mm1
         u(i,n)=0.0D+0
      END DO

      CALL VRFFTB(mm1,n,u(1,1),fftwork,sizep1,wsave)      
C               
C set BC and periodicity, then check L-infinity error     
C           
      DO j=1,n
         u(0,j)=0.0D+0
         u(m,j)=0.0D+0
      END DO
      DO i=0,m
         u(i,0)=u(i,n)
      END DO
               
      t1=0.0D+0
      DO j=0,n
         DO i=0,m
            t2=DABS(u(i,j)-tu(i,j))
            IF (t2 .GT. t1) THEN
               t1=t2
            END IF
         END DO
      END DO
      
      WRITE(6,*) 'L-inf error : ',t1      

      STOP
      END
      
!!!!!!!!!!!!!!!!!!!!!

      SUBROUTINE CHEB(x,D1,D2,m,arrdim)
      IMPLICIT NONE 
      
C input/output variables
     
      INTEGER m,arrdim
      REAL*8  x(arrdim+1),D1(arrdim+1,arrdim+1),D2(arrdim+1,arrdim+1)

C local parameters

      INTEGER size
      PARAMETER(size=65)  
      REAL*8 pi
      PARAMETER(pi=3.14159265358979323846264338327950288D+0)

C local variables           
           
      INTEGER i,j,k,n,n1,n2
      REAL*8  th(size),z(size,size),dx(size,size)
      REAL*8  cvec(size),c(size,size),sum
      
C executable code
      
      IF (size .LT. m+1) THEN
         WRITE(6,*) 'In CHEB: SIZE is too small!!'
         WRITE(6,*) 'Must be at least ',m+1
         STOP
      END IF    
       
      n=m+1

      j=m
      DO i=1,n
         th(i)=DBLE(i-1)*pi/DBLE(m)
         x(i)=DSIN(DBLE(j)*pi/DBLE(2*m))
         j=j-2
      END DO     

      DO i=1,n
         DO j=1,n
            z(i,j)=2.0D+0*DSIN(th(j)/2.0D+0+th(i)/2.0D+0)*
     &                    DSIN(th(j)/2.0D+0-th(i)/2.0D+0)
         END DO
      END DO

      n1=n/2
      n2=n-n1 
      DO j=1,n
         DO i=1,n1
            dx(i,j)=z(i,j)
         END DO
         DO i=1,n2
            dx(n1+i,j)=-z(n2+1-i,n+1-j)
         END DO
         dx(j,j)=1.0D+0
      END DO         
      DO j=1,n
         DO i=1,n
            z(i,j)=1.0D+0/dx(i,j)
         END DO
         z(j,j)=0.0D+0
      END DO
   
      cvec(1)=1.0D+0
      DO i=2,n
         cvec(i)=-cvec(i-1)
      END DO
      DO i=1,n
         k=1
         DO j=i,n
            c(i,j)=cvec(k)
            k=k+1
         END DO 
         k=2
         DO j=i-1,1,-1
            c(i,j)=cvec(k)
            k=k+1
         END DO
      END DO 
      DO i=2,n-1
         c(i,1)=c(i,1)/2.0D+0
         c(i,n)=c(i,n)/2.0D+0
      END DO       
      DO j=2,n-1
         c(1,j)=2.0D+0*c(1,j)
         c(n,j)=2.0D+0*c(n,j)
      END DO
 
      
C compute D1

      DO j=1,n
         DO i=1,n
            D1(i,j)=z(i,j)*c(i,j)
         END DO
         D1(j,j)=z(j,j)*(c(j,j)-1.0D+0)
      END DO
      DO i=1,n
         sum=0.0D+0
         DO j=1,n
            sum=sum+D1(i,j)
         END DO
         D1(i,i)=-sum
      END DO

C compute D2

      DO j=1,n
         DO i=1,n
            D2(i,j)=2.0D+0*z(i,j)*(c(i,j)*D1(i,i)-D1(i,j))
         END DO
      END DO
      DO i=1,n
         sum=0.0D+0
         DO j=1,n
            sum=sum+D2(i,j)
         END DO
         D2(i,i)=-sum
      END DO

      RETURN
      END
      
